/**
 * @file   tclsqlitevfs.c
 * @author Anton Kovalenko <anton@sw4me.com>
 * @date   Tue Feb 23 02:51:17 2010
 * 
 * @brief  sqlite3-tcl vfs bridge. pre-alpha version.
 * (c) 2010 Anton Kovalenko, unreleased draft.
 * 
 */

#include <tcl.h>

#include <sqlite3.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>


struct subclassed_tcl_file {
    struct sqlite3_file sqlite3File;
    Tcl_Channel chan;
};
TCL_DECLARE_MUTEX(vfsRegistrationLock);

static struct sqlite3_vfs tcl_sqlite3_vfs;
static struct sqlite3_vfs *former_sqlite3_vfs;

int Tclsqlitevfs_Init(Tcl_Interp *interp)
{
    int result = TCL_OK;
    if (!Tcl_InitStubs(interp,"8.4", 0))
	return TCL_ERROR;

    Tcl_MutexLock(&vfsRegistrationLock);
    {
	if (!former_sqlite3_vfs) {
	    sqlite3_vfs* oldVfs = sqlite3_vfs_find(NULL);
	    if (oldVfs == NULL) {
		Tcl_AppendResult(interp,
				 "Can't find original (e.g. native) sqlite3 vfs",
				 NULL);
		result = TCL_ERROR;
		goto end;
	    }
	    former_sqlite3_vfs = oldVfs;
	    tcl_sqlite3_vfs.szOsFile = 
		oldVfs->szOsFile > (sizeof (struct subclassed_tcl_file) )?
		oldVfs->szOsFile : (sizeof (struct subclassed_tcl_file) );
	    sqlite3_vfs_register(&tcl_sqlite3_vfs, 1); /* register and make default */
	}
    }
end:
    Tcl_MutexUnlock(&vfsRegistrationLock);
    return result;
}


static int tsq_xOpen(sqlite3_vfs* vfs, const char *zName, sqlite3_file* pFile,
		     int flags, int *pOutFlags);

static int tsq_xDelete(sqlite3_vfs* vfs, const char *zName, int syncDir);

static int tsq_xAccess(sqlite3_vfs* vfs, const char *zName, int flags, int *pResOut);

static int tsq_xFullPathname(sqlite3_vfs* vfs, const char *zName, int nOut, char *zOut);

static void *tsq_xDlOpen(sqlite3_vfs* vfs, const char *zName);

static void tsq_xDlError(sqlite3_vfs* vfs, int nByte, char *zErrMsg);

static void (*tsq_xDlSym(sqlite3_vfs* vfs, void* handle, const char *zSymbol))(void);

static void tsq_xDlClose(sqlite3_vfs* vfs, void* handle);

static int tsq_xRandomness(sqlite3_vfs* vfs, int nByte, char *zOut);

static int tsq_xSleep(sqlite3_vfs* vfs, int microseconds);

static int tsq_xCurrentTime(sqlite3_vfs* vfs, double* timePtr);

static int tsq_xGetLastError(sqlite3_vfs* vfs, int nChar, char* msg);

static struct sqlite3_vfs tcl_sqlite3_vfs = {
    .iVersion = 1,
    .szOsFile = 0,
    .mxPathname = 4200,
    .pNext = NULL,
    .zName = "tcl_sqlite3_vfs",
    .pAppData = NULL,
    .xOpen = &tsq_xOpen,
    .xDelete = &tsq_xDelete,
    .xAccess = &tsq_xAccess,
    .xFullPathname = &tsq_xFullPathname,
    .xDlOpen = &tsq_xDlOpen,
    .xDlError = &tsq_xDlError,
    .xDlSym = &tsq_xDlSym,
    .xDlClose = &tsq_xDlClose,
    .xRandomness = &tsq_xRandomness,
    .xSleep = &tsq_xSleep,
    .xCurrentTime = &tsq_xCurrentTime,
    .xGetLastError = &tsq_xGetLastError,
};


static int tclfile_xClose(struct sqlite3_file* fptr);
static int tclfile_xRead(struct sqlite3_file* fptr, void* dst, int iAmt,
			 sqlite3_int64 iOfst);
static int tclfile_xWrite(struct sqlite3_file* fptr, const void* src, int iAmt,
			  sqlite3_int64 iOfst);
static int tclfile_xTruncate(struct sqlite3_file* fptr, sqlite3_int64 size);
static int tclfile_xSync(struct sqlite3_file* fptr, int flags);
static int tclfile_xFileSize(struct sqlite3_file* fptr, sqlite3_int64 *pSize);

static int tclfile_xLock(struct sqlite3_file* fptr, int flags);
static int tclfile_xUnlock(struct sqlite3_file* fptr, int flags);
static int tclfile_xCheckReservedLock(struct sqlite3_file* fptr, int *pOut);
static int tclfile_xFileControl(struct sqlite3_file* fptr, int op, void *pArg);
static int tclfile_xSectorSize(struct sqlite3_file* fptr);
static int tclfile_xDeviceCharacteristics(struct sqlite3_file* fptr);

const static struct sqlite3_io_methods tcl_io_methods = {
    .iVersion = 1,
    .xClose = &tclfile_xClose,
    .xRead = &tclfile_xRead,
    .xWrite = &tclfile_xWrite,
    .xTruncate = &tclfile_xTruncate,
    .xSync = &tclfile_xSync,
    .xFileSize = &tclfile_xFileSize,
    .xLock = &tclfile_xLock,
    .xUnlock = &tclfile_xUnlock,
    .xCheckReservedLock = &tclfile_xCheckReservedLock,
    .xFileControl = &tclfile_xFileControl,
    .xSectorSize = &tclfile_xSectorSize,
    .xDeviceCharacteristics = &tclfile_xDeviceCharacteristics,
};



static int tsq_xRandomness(sqlite3_vfs* vfs, int nByte, char *zOut)
{
    return
	former_sqlite3_vfs->
	xRandomness(former_sqlite3_vfs, nByte, zOut);
}

static int tsq_xSleep(sqlite3_vfs* vfs, int microseconds)
{
    return
	former_sqlite3_vfs->
	xSleep(former_sqlite3_vfs, microseconds);
}

static int tsq_xCurrentTime(sqlite3_vfs* vfs, double* timePtr) 
{
    return
	former_sqlite3_vfs->
	xCurrentTime(former_sqlite3_vfs, timePtr);
}

static int tsq_xGetLastError(sqlite3_vfs* vfs, int nChar, char* msg)
{
    return
	former_sqlite3_vfs->
	xGetLastError(former_sqlite3_vfs, nChar, msg);
}

static void *tsq_xDlOpen(sqlite3_vfs* vfs, const char *zName)
{
    return
	former_sqlite3_vfs->
	xDlOpen(former_sqlite3_vfs, zName);

}

static void tsq_xDlError(sqlite3_vfs* vfs, int nByte, char *zErrMsg)
{
    former_sqlite3_vfs->
	xDlError(former_sqlite3_vfs, nByte, zErrMsg);
}

static void (*tsq_xDlSym(sqlite3_vfs* vfs, void* handle, const char *zSymbol))(void)
{
    return
	former_sqlite3_vfs->
	xDlSym(former_sqlite3_vfs, handle, zSymbol);
}

static void tsq_xDlClose(sqlite3_vfs* vfs, void* handle)
{
    former_sqlite3_vfs->
	xDlClose(former_sqlite3_vfs,handle);
}

static int tsq_xOpen(sqlite3_vfs* vfs, const char *zName, sqlite3_file* pFile,
	      int flags, int *pOutFlags)
{
    int rc = SQLITE_OK;
    Tcl_Obj *pathObj = Tcl_NewStringObj(zName,-1);
    Tcl_IncrRefCount(pathObj);
    {
	const Tcl_Filesystem* fs = Tcl_FSGetFileSystemForPath(pathObj);
	if (!strcmp(fs->typeName,"native")) {
	    Tcl_DecrRefCount(pathObj);
	    return former_sqlite3_vfs->xOpen(former_sqlite3_vfs, zName, 
					     pFile, flags, pOutFlags);
	}
	/* It is the VFS path, so we'll open it like that */
	pFile->pMethods = NULL;
	{
	    struct subclassed_tcl_file *f =
		(struct subclassed_tcl_file*) pFile;
	    const char* modeString;
	    int permissions = 0640;
	    
	    if (flags & SQLITE_OPEN_READWRITE) {
		if (flags & SQLITE_OPEN_CREATE) {
		    if (flags & SQLITE_OPEN_EXCLUSIVE) {
			modeString = "CREAT EXCL RDWR";
		    } else {
			modeString = "CREAT RDWR";
		    }
		} else {
		    modeString = "RDWR";
		}
	    } else {
		modeString = "RDONLY";
	    }
	    Tcl_Channel chan = Tcl_FSOpenFileChannel(NULL,
						     pathObj,
						     modeString,
						     permissions);
	    if ((chan == NULL) && 
		(flags & SQLITE_OPEN_READWRITE)) {
		/* fall back to read-only mode */
		chan = Tcl_FSOpenFileChannel(NULL,
					     pathObj,
					     "r",
					     permissions);
	    }
	    if (chan) {
		Tcl_SetChannelOption(NULL,chan,"-translation","binary");
		Tcl_SetChannelOption(NULL,chan,"-encoding","binary");
		Tcl_SetChannelOption(NULL,chan,"-buffering","none");
		Tcl_SetChannelOption(NULL,chan,"-eofchar","");
		f->chan = chan;
		f->sqlite3File.pMethods = &tcl_io_methods;
	    } else {
		rc = SQLITE_CANTOPEN;
	    }
	}
    }
    Tcl_DecrRefCount(pathObj);
    return rc;
}

static int tsq_xDelete(sqlite3_vfs* vfs, const char *zName, int syncDir)
{
    Tcl_Obj *pathObj = Tcl_NewStringObj(zName,-1);
    int rc = SQLITE_OK;

    Tcl_IncrRefCount(pathObj);
    {
	const Tcl_Filesystem* fs = Tcl_FSGetFileSystemForPath(pathObj);
	if (!strcmp(fs->typeName,"native")) {
	    Tcl_DecrRefCount(pathObj);
	    return former_sqlite3_vfs->xDelete(former_sqlite3_vfs, zName, 
					       syncDir);
	}
	/* It is the VFS path, so we'll delete it with TCL */
	/* FIXME: don't know how to sync dir though, so just ignore it :( */
	rc = Tcl_FSDeleteFile(pathObj) == TCL_OK ? 
	    SQLITE_OK : SQLITE_IOERR_DELETE;
    }
    Tcl_DecrRefCount(pathObj);
    return rc;
}



static int tsq_xAccess(sqlite3_vfs* vfs, const char *zName, int flags, int *pResOut)
{
    int amode = 0;
    Tcl_Obj *pathObj = Tcl_NewStringObj(zName,-1);
    int rc = SQLITE_OK;

    Tcl_IncrRefCount(pathObj);

    switch( flags ){
    case SQLITE_ACCESS_EXISTS:
	amode = F_OK;
	break;
    case SQLITE_ACCESS_READWRITE:
	amode = W_OK|R_OK;
	break;
    case SQLITE_ACCESS_READ:
	amode = R_OK;
	break;

    default:
	Tcl_Panic("Invalid flags argument");
    }
    
    *pResOut = (Tcl_FSAccess(pathObj, amode)==0);
    Tcl_DecrRefCount(pathObj);
    return rc;
}

static int tsq_xFullPathname(sqlite3_vfs* vfs, const char *zName, int nOut, char *zOut)
{
    int amode = 0;
    Tcl_Obj *pathObj = Tcl_NewStringObj(zName,-1);
    int rc = SQLITE_OK;

    Tcl_IncrRefCount(pathObj);
    {
	Tcl_Obj* normalizedPath = Tcl_FSGetNormalizedPath(NULL,pathObj);
	if (normalizedPath) {
	    int length = 0;
	    char* strNormalizedPath = 
		Tcl_GetStringFromObj(normalizedPath, &length);
	    if (length > nOut - 1) {
		rc = SQLITE_CANTOPEN;
	    } else {
		memcpy(zOut, strNormalizedPath, length);
		zOut[length]='\000';
	    }
	} else {
	    rc = SQLITE_CANTOPEN;
	}
    }
    return rc;
}

static int tclfile_xClose(struct sqlite3_file* fptr)
{
    struct subclassed_tcl_file* f = 
	(struct subclassed_tcl_file*)fptr;
    return Tcl_Close(NULL, f->chan) == TCL_OK ? 
	SQLITE_OK : SQLITE_IOERR_CLOSE;
}

static int tclfile_xRead(struct sqlite3_file* fptr,
			 void* dst, int iAmt,
			 sqlite3_int64 iOfst)
{
    struct subclassed_tcl_file* f = 
	(struct subclassed_tcl_file*)fptr;
    Tcl_WideInt newPos = Tcl_Seek(f->chan, iOfst, SEEK_SET);

    int hasRead = Tcl_ReadRaw(f->chan, dst, iAmt);
    if (hasRead < 0) {
	return SQLITE_IOERR_READ;
    }
    if (hasRead < iAmt) {
	memset(dst+hasRead,0,iAmt - hasRead);
	return SQLITE_IOERR_SHORT_READ;
    }
    return SQLITE_OK;
}

static int tclfile_xWrite(struct sqlite3_file* fptr,
			  const void* src, int iAmt,
			  sqlite3_int64 iOfst)
{
    struct subclassed_tcl_file* f = 
	(struct subclassed_tcl_file*)fptr;
    Tcl_WideInt newPos = Tcl_Seek(f->chan, iOfst, SEEK_SET);
    int hasWritten = Tcl_Write(f->chan, src, iAmt);
    if (hasWritten < iAmt) {
	return SQLITE_IOERR_WRITE;
    } 
    return SQLITE_OK;
}

static int tclfile_xTruncate(struct sqlite3_file* fptr,
			     sqlite_int64 size)
{
    struct subclassed_tcl_file* f = 
	(struct subclassed_tcl_file*)fptr;
    return Tcl_TruncateChannel(f->chan, size) == TCL_OK ?
	SQLITE_OK : SQLITE_IOERR_TRUNCATE;
}

static int tclfile_xSync(struct sqlite3_file* fptr, int flags)
{
    return SQLITE_OK;
}

static int tclfile_xFileSize(struct sqlite3_file* fptr, sqlite3_int64 *pSize)
{
    struct subclassed_tcl_file* f = 
	(struct subclassed_tcl_file*)fptr;
    *pSize = Tcl_Seek(f->chan,0, SEEK_END);
    return SQLITE_OK;
}

static int tclfile_xLock(struct sqlite3_file* fptr, int flags)
{
    return SQLITE_OK;
}
static int tclfile_xUnlock(struct sqlite3_file* fptr, int flags)
{
    return SQLITE_OK;
}
static int tclfile_xCheckReservedLock(struct sqlite3_file* fptr, int *pOut)
{
    *pOut = 0;
    return SQLITE_OK;
}
static int tclfile_xFileControl(struct sqlite3_file* fptr, int op, void *pArg)
{
    return SQLITE_OK;
}
static int tclfile_xSectorSize(struct sqlite3_file* fptr)
{
    return 512;
}
static int tclfile_xDeviceCharacteristics(struct sqlite3_file* fptr)
{
    return 0;
}
